--[[
Name: LibNameplate-1.0
Author(s): Cyprias (cyprias@gmail.com)
Documentation: http://www.wowace.com/addons/libnameplate-1-0/pages/main/
SVN:  svn://svn.wowace.com/wow/libnameplate-1-0/mainline/trunk
Description: Alerts addons when a nameplate is shown or hidden. Has API to get info such as name, level, class, ect from the nameplate. LibNameplate tries to function with the default nameplates, Aloft, caelNamePlates and TidyPlates (buggy).
Dependencies: LibStub, CallbackHandler-1.0
]]

--[[
	Todo
		- confirm GetThreat() works with aloft.
	Notes
		- hooking plates OnEnter requires the plates enable mouse. But doing that stops the mouseover ID being available. 
]]

local MAJOR, MINOR = "LibNameplate-1.0", 6
if not LibStub then error(MAJOR .. " requires LibStub.") return end

local lib = LibStub:NewLibrary(MAJOR, MINOR)
if not lib then return end

lib.callbacks = lib.callbacks or LibStub("CallbackHandler-1.0"):New(lib)
if not lib.callbacks then	error(MAJOR .. " CallbackHandler-1.0.") return end

-- Globals
local _G					= _G
local tostring				= tostring
local WorldFrame			= WorldFrame
local select				= select
local UnitExists			= UnitExists
local UnitGUID				= UnitGUID
local UnitName				= UnitName
local string_gsub			= string.gsub
local pairs					= pairs
local table_remove			= table.remove
local table_insert			= table.insert
local GetMouseFocus			= GetMouseFocus
local GetNumRaidMembers		= GetNumRaidMembers
local GetNumPartyMembers	= GetNumPartyMembers
local UnitIsUnit			= UnitIsUnit
local GetRaidTargetIndex	= GetRaidTargetIndex
local UnitInParty			= UnitInParty
local tonumber				= tonumber
local print					= print
local math_floor			= math.floor
local RAID_CLASS_COLORS		= RAID_CLASS_COLORS
local table_getn			= table.getn
local CreateFrame			= CreateFrame
local _ --localized underscore so FindGlobals doesn't bug me about it.
local IsAddOnLoaded			= IsAddOnLoaded
local type = type


--Conts
local scanDelay		= 1 -- Scan for new nameplates every 1 seconds. We hook OnShow so this doesn't need to be done rapidly.
local updateDelay	= 1 -- Update nameplate info every 1 seconds. This checks if a nameplate has a raid icon then trys to find the GUID to that icon.

--these aren't really used. more to help me debug for now.
local aloftRunning			= false
local tidyPlatesRunning 	= false
local caelNPRunning			= false

local DEBUG = false
--[===[@debug@
DEBUG = true
--@end-debug@]===]

local realPlate = {}
local fakePlate = {}

local function CmdHandler()
	DEBUG = not DEBUG
end
_G.SlashCmdList["LIBNAMEPLATEDEBUG"] = CmdHandler
_G.SLASH_LIBNAMEPLATEDEBUG1 = "/lnbug"
local function debugPrint(...)
	if DEBUG then
		print(...)
	end
end

lib.nameplates		= lib.nameplates or {}
lib.GUIDs			= lib.GUIDs or {}
lib.names			= lib.names or {}
lib.onShowHooks = lib.onShowHooks or {}
lib.onHideHooks = lib.onHideHooks or {}
lib.onUpdateHooks = lib.onUpdateHooks or {}
lib.plateGUIDs = lib.plateGUIDs or {}
lib.isOnScreen = lib.isOnScreen or {}

--[[
TidyPlate region names
	old: "threatGlow", "healthBorder", "castBorder", "castNostop", "spellIcon", "highlightTexture", "nameText", "levelText", "dangerSkull", "raidIcon", "eliteIcon"
	new: "threatGlow", "healthborder", "castborder", "castnostop", "spellicon", "highlight", "name", "level", "dangerskull", "raidicon", "eliteicon"
TidyPlate child names
	old: "healthBar", "castBar"
	new: "castbar", "healthbar"
]]
local regionNames = { "threatGlow", "healthBorder", "castBorder", "castNostop", "spellIcon", "highlightTexture", "nameText", "levelText", "dangerSkull", "raidIcon", "eliteIcon" }
local regions = {}
for i, name in pairs(regionNames) do 
	regions[name] = i
end

local barNames = { "healthBar", "castBar"}
local bars = {}
for i, name in pairs(barNames) do 
	bars[name] = i
end



--"Star", "Circle", "Diamond", "Triangle", "Moon", "Square", "Cross", "Skull"
--~ local raidIconTexture = "Interface\\TargetingFrame\\UI-RaidTargetingIcons"
--			= left							right
--			= ULx	ULy		LLx		LLy		URx		URy		LRx		LRy
-- star		= 0		0		0		0.25	0.25	0		0.25	0.25
-- circle	= 0.25	0		0.25	0.25	0.5		0		0.5		0.25
-- diamond	= 0.5	0		0.5		0.25	0.75	0		0.75	0.25
-- triangle	= 0.75	0		0.75	0.25	1		0		1		0.25
-- moon		= 0		0.25	0		0.5		0.25	0.25	0.25	0.5
-- square 	= 0.25	0.25	0.25	0.5		0.5		0.25	0.5		0.5
-- cross	= 0.5	0.25	0.5		0.5		0.75	0.25	0.75	0.5
-- skull	= 0.75	0.25	0.75	0.5		1		0.25	1		0.5
local raidIconTexCoord = {--from GetTexCoord. input is ULx and ULy (first 2 values).
	[0]		= {
		[0]		= 1, --star 
		[0.25]	= 5, --moon
	},
	[0.25]	= {
		[0]		= 2, --circle 
		[0.25]	= 6, --square
	},
	[0.5]	= {
		[0]		= 3, --star 
		[0.25]	= 7, --cross
	},
	[0.75]	= {
		[0]		= 4, --star 
		[0.25]	= 8, --skull
	},
}



local colorToClass = {}
local function pctToInt(number) return math_floor((100*number) + 0.5) end
for classname, color in pairs(RAID_CLASS_COLORS) do
	colorToClass["C"..pctToInt(color.r)+pctToInt(color.g)+pctToInt(color.b)] = classname
end


--[[
	these always return nil
	frame:IsClampedToScreen(), frame:IsIgnoringDepth(), frame:IsMovable(), frame:IsToplevel(), frame:GetDepth(), frame:GetEffectiveDepth(), 

]]
local function IsNamePlateFrame(frame)
	if frame:GetName() then
--~ 		debugPrint("IsNamePlateFrame","No name!")
		return false
	end

	if frame:GetID() ~= 0 then
		return false
	end
	
	if frame:GetObjectType() ~= "Frame" then
--~ 		debugPrint("IsNamePlateFrame","Not a frame!")
		return false
	end
		
	if frame.extended or frame.aloftData then
		--Tidyplates = extended, Aloft = aloftData
		--They sometimes remove & replace the children so this needs to be checked first.
		return true
	end

	if frame.done then --caelNP
		return true
	end
	
	if frame:GetNumChildren() == 0 then
--~ 		debugPrint("IsNamePlateFrame","No children?!!", frame:GetName(), frame.extended and frame.extended.bars)
		return false
	end
		
	if frame:GetNumRegions() == 0 then
		return false
	end
	
	return true
end



--[===[@debug@
function LC_TEST()--		/run LC_TEST()
	local c = {WorldFrame:GetChildren()}
	local frame, cFrame
	for i = 1, table_getn(c) do
		frame = c[i]
		if frame:IsShown() then
			--frame.extended
			debugPrint(IsNamePlateFrame(frame), "ID:"..tostring(frame:GetID()), "C:"..tostring(frame:GetNumChildren()), "R:"..tostring(frame:GetNumRegions()))
			
		end
	end
end
--@end-debug@]===]

local function ScanWorldFrameChildren(n, ...)
    for i = 1, n do
        local frame = select(i, ...)
        if frame:IsShown() and IsNamePlateFrame(frame) then
            if not lib.nameplates[frame] then
                lib:SetupNameplate(frame)                    
            end
        end    
    end
end


local lastChildren = 0
local function FindNameplates()
    local curChildren = WorldFrame:GetNumChildren()
    if curChildren ~= lastChildren then
        lastChildren = curChildren
        ScanWorldFrameChildren(curChildren, WorldFrame:GetChildren())
    end
end

local function threatByColor( object , name)
	local redCan, greenCan, blueCan, alphaCan = object:GetVertexColor()
	local shown
	if not object:IsShown() then return "LOW" end
	if greenCan > .7 then return "MEDIUM" end
	if redCan > .7 then return "HIGH" end
end

local function reactionByColor(red, green, blue, a)																											
	if red < .01 and blue < .01 and green > .99 then return "FRIENDLY", "NPC" 
	elseif red < .01 and blue > .99 and green < .01 then return "FRIENDLY", "PLAYER"
	elseif red > .99 and blue < .01 and green > .99 then return "NEUTRAL", "NPC"
	elseif red > .99 and blue < .01 and green < .01 then return "HOSTILE", "NPC"
	else return "HOSTILE", "PLAYER" end
end

local function combatByColor(r, g, b, a) 
	return (r > .5 and g < .5)
end

function lib:GetName(frame)
	local frame = realPlate[frame] or frame
	if frame.extended then
		if frame.extended.unit then
			if frame.extended.unit.name then
--~ 				debugPrint("GetName 7","name in unit")
				return frame.extended.unit.name
			end
		end
		
		if frame.extended.regions then
			if frame.extended.regions.name then
				if type(frame.extended.regions.name) ~= "table" then
					local name = string_gsub(frame.extended.regions.name, "|cffffffff", "")
					return name
				else
					if frame.extended:GetNumRegions() > 0 then
						local regs = {frame.extended:GetRegions()}
						
						if regs[13] and regs[13].GetText and type(regs[13]:GetText()) == "string" then
--~ 							debugPrint("GetName 9", "Found name #13", regs[13]:GetText())
							return regs[13]:GetText()
						end

					--[[for i=1, table.getn(regs) do 
							if regs[i].GetText and regs[i]:GetText() and type(regs[i]:GetText()) == "string" then
								debugPrint("GetName 10", i, regs[i], regs[i]:GetText())
							end
						end]]
					end
				end
			end
		end

	elseif frame.aloftData and frame.aloftData.nameTextRegion then
		local name = string_gsub(frame.aloftData.nameTextRegion:GetText(), "|cffffffff", "")
	end

	--string_gsub returns 2 values. I just want to return 1.
	local name = string_gsub(select(regions.nameText, frame:GetRegions()):GetText(), "|cffffffff", "")
	return name
end

function lib:GetScale(frame)
	local frame = realPlate[frame] or frame
	if frame.extended then
		frame = frame.extended
	end
	return frame:GetScale()
end

function lib:GetVisibleFrame(frame)
	local frame = realPlate[frame] or frame
	if frame.extended then
		frame = frame.extended
	end
	return frame
end

function lib:GetLevel(frame)
	local frame = realPlate[frame] or frame
	if frame.extended then
		return tonumber(frame.extended.unit.level or 0)
	elseif frame.aloftData then
		return tonumber(frame.aloftData.level or 0)
	end
	return tonumber(select(regions.levelText, frame:GetRegions()):GetText() or 0)
end

function lib:GetReaction(frame)
	local frame = realPlate[frame] or frame
	if frame.extended then
		if frame.extended.unit and frame.extended.unit.reaction then
			return frame.extended.unit.reaction
		end
		
		if frame.extended.bars then
			if frame.extended.bars.health and frame.extended.bars.health.GetStatusBarColor then
				return reactionByColor(frame.extended.bars.health:GetStatusBarColor())
			end
		end

		if not select(bars.healthBar, frame:GetChildren()).GetStatusBarColor then
			return "UNKNOWN"
		end
		
	elseif frame.aloftData then
		local r, g, b = frame.aloftData.originalHealthBarR, frame.aloftData.originalHealthBarG, frame.aloftData.originalHealthBarB
		return reactionByColor(r, g, b, 1)
	end
	return reactionByColor(select(bars.healthBar, frame:GetChildren()):GetStatusBarColor())
end

function lib:GetType(frame)
	local frame = realPlate[frame] or frame
	if frame.extended then
		if frame.extended.unit and frame.extended.unit.type then
			return frame.extended.unit.type
		end

		if frame.extended.bars then
			if frame.extended.bars.health and frame.extended.bars.health.GetStatusBarColor then
				return select(2, reactionByColor(frame.extended.bars.health:GetStatusBarColor()))
			end
		end

		if not select(bars.healthBar, frame:GetChildren()).GetStatusBarColor then
			return "UNKNOWN"
		end
	end

	return select(2, reactionByColor(select(bars.healthBar, frame:GetChildren()):GetStatusBarColor()))
end

function lib:IsBoss(frame)
	local frame = realPlate[frame] or frame
	if frame.extended then
		return frame.extended.unit.isBoss
	elseif frame.aloftData and frame.aloftData.bossIconRegion then
		return frame.aloftData.bossIconRegion:IsShown()
	end
	return select(regions.dangerSkull, frame:GetRegions()):IsShown()
end

function lib:GetHealthMax(frame)
	local frame = realPlate[frame] or frame
	if frame.extended then
		return tonumber(frame.extended.unit.healthmax or 0)
	end
	local _, max = select(bars.healthBar, frame:GetChildren()):GetMinMaxValues()
	return tonumber(max or 0)
end

--This will return nil if we're not in a PvP zone (like in cities)
function lib:GetClass(frame)
	local frame = realPlate[frame] or frame
	if frame.extended then
		return frame.extended.unit.class
	end
	local red, green, blue = select(bars.healthBar, frame:GetChildren()):GetStatusBarColor()
	return colorToClass["C"..pctToInt(red)+pctToInt(green)+pctToInt(blue)]
end

function lib:IsElite(frame)
	local frame = realPlate[frame] or frame
	if frame.extended then
		return frame.extended.unit.isElite
	elseif frame.aloftData and frame.aloftData.stateIconRegion then
		return frame.aloftData.stateIconRegion:IsShown()
	end
	return select(regions.eliteIcon, frame:GetRegions()):IsShown()
end

function lib:GetThreatSituation(frame)
	local frame = realPlate[frame] or frame
	if frame.extended then
		return frame.extended.unit.threatSituation
	elseif frame.aloftData and frame.aloftData.nativeGlowRegion then
		return threatByColor(frame.aloftData.nativeGlowRegion)
	end
	return threatByColor(select(regions.threatGlow, frame:GetRegions()))
end

function lib:IsTarget(frame)
	local frame = realPlate[frame] or frame
	return frame:IsShown() and frame:GetAlpha() == 1 and UnitExists("target")
end

function lib:GetHealth(frame)
	local frame = realPlate[frame] or frame
	if frame.extended then
		return tonumber(frame.extended.unit.health or 0)
	end
	return tonumber(select(bars.healthBar, frame:GetChildren()):GetValue() or 0)
end

------------------------------------------------------------------------------------------------------------------
local function HideMouseoverRegion(frame)																		--
-- If we move the camera angle while the mouse is over a plate, that plate won't hide the mouseover texture.	--
-- So if we're mousing over someone's feet and a plate has the mouseover texture visible, 						--
-- it fools our code into thinking we're mousing over that plate.												--
-- This can be recreated by placing the mouse over a nameplate then holding rightclick and moving the canera.	--
-- Our UpdateNameplateInfo will see that the texture is still visible but no mouseover ID is available. 		--
------------------------------------------------------------------------------------------------------------------
	if frame.extended then
		if frame.extended.regions.highlightTexture then --old TidyPlate
			frame.extended.regions.highlightTexture:Hide()
		elseif frame.extended.regions.highlight then --new TidyPlate
			frame.extended.regions.highlight:Hide()
		end
		return
	elseif frame.aloftData and frame.aloftData.highlightRegion then
		frame.aloftData.highlightRegion:Hide()
--~ 		debugPrint("Aloft's running, hiding it's highlight region.", lib:GetName(frame), lib:GetLevel(frame))
		return
	end
	return select(regions.highlightTexture, frame:GetRegions()):Hide()
end

function lib:IsMouseover(frame)
	local frame = realPlate[frame] or frame
	if frame.extended then
--~ 		return frame.extended.unit.isMouseover
		if frame.extended.regions.highlightTexture then --old TidyPlate
			return frame.extended.regions.highlightTexture:IsShown() --better then checking if TidyPlates has saved isMosueover var.
		elseif frame.extended.regions.highlight then --new TidyPlate
			return frame.extended.regions.highlight:IsShown() --better then checking if TidyPlates has saved isMosueover var.
		end
	elseif frame.aloftData then
		if frame.aloftData.highlightRegion then
			return frame.aloftData.highlightRegion:IsShown()
		end
	end
	return select(regions.highlightTexture, frame:GetRegions()):IsShown()
end


function lib:IsCasting(frame)
	local frame = realPlate[frame] or frame
	if frame.extended then
		return frame.extended.unit.isCasting
	end
	return select(bars.castBar, frame:GetChildren()):IsShown()
end

function lib:IsInCombat(frame)
	local frame = realPlate[frame] or frame
	if frame.extended then
		return frame.extended.unit.isInCombat
	end
	return combatByColor(select(regions.nameText, frame:GetRegions()):GetTextColor())
end

function lib:IsMarked(frame)
	local frame = realPlate[frame] or frame
	if frame.extended then
		return frame.extended.unit.isMarked
	elseif frame.aloftData and frame.aloftData.raidIconRegion then
		return frame.aloftData.raidIconRegion:IsShown()
	end
	return select(regions.raidIcon, frame:GetRegions()):IsShown()
end

function lib:GetRaidIcon(frame)
	local frame = realPlate[frame] or frame
	local raidIcon
	if frame.extended then
		if frame.extended.regions.raidIcon then --old TidyPlate
			raidIcon = frame.extended.regions.raidIcon
		elseif frame.extended.regions.raidicon then --new TidyPlate
			raidIcon = frame.extended.regions.raidicon
		else
			return 0
		end
	elseif frame.aloftData and frame.aloftData.raidIconRegion then
		raidIcon = frame.aloftData.raidIconRegion
	else
		raidIcon = select(regions.raidIcon, frame:GetRegions())
	end

	if raidIcon:IsShown() then
		local ULx,ULy = raidIcon:GetTexCoord() --,LLx,LLy,URx,URy,LRx,LRy
		if ULx and ULy then
			return raidIconTexCoord[ULx] and raidIconTexCoord[ULx][ULy] or 0
		end
	end
	return 0
end

function lib:GetGUID(frame)
	local frame = realPlate[frame] or frame
	return self.plateGUIDs[frame]
end

-- ***
function lib:GetTargetNameplate()
	if self.targeted and self.targeted:IsShown() then
		return fakePlate[self.targeted] or self.targeted
	end
end

function lib:GetNameplateByGUID(GUID)
	if self.GUIDs[GUID] and self.GUIDs[GUID]:IsShown() then
		return fakePlate[self.GUIDs[GUID]] or self.GUIDs[GUID]
	end
end


function lib:GetNameplateByName(name)
	if self.names[name] and self.names[name]:IsShown() then
		return fakePlate[self.names[name]] or self.names[name]
	end
end

function lib:GetNameplateByUnit(unitID)
	if UnitIsUnit(unitID, "target") then
		return self:GetTargetNameplate()
	end
	local GUID = UnitGUID(unitID)
	if self.GUIDs[GUID] and self.GUIDs[GUID]:IsShown() then
		return fakePlate[self.GUIDs[GUID]] or self.GUIDs[GUID]
	end
	local name = UnitName(unitID)
	if self.names[GUID] and self.names[GUID]:IsShown() then
		return fakePlate[self.names[GUID]] or self.names[GUID]
	end
end


local function RecycleNameplate(frame)
--~ 	debugPrint("RecycleNameplate", frame)

	lib.callbacks:Fire("LibNameplate_RecycleNameplate", fakePlate[frame] or frame)
	
	if lib.plateGUIDs[frame] then
		lib.GUIDs[lib.plateGUIDs[frame]] = false
	end
	
	if fakePlate[frame] then
		lib.names[fakePlate[frame]] = false
	end
	
	
	
	lib.plateGUIDs[frame] = false
	local fake = fakePlate[frame]
	if fake then
		realPlate[fake] = false
	end
	fakePlate[frame] = false
end

local function FoundPlateGUID(frame, GUID, unitID)
	lib.plateGUIDs[frame] = GUID
	
	lib.GUIDs[GUID] = frame
	lib.callbacks:Fire("LibNameplate_FoundGUID", fakePlate[frame] or frame, GUID, unitID)
end

local function GetMouseoverGUID(frame)
	local unitID = "mouseover"
	if UnitExists(unitID) then
		FoundPlateGUID(frame, UnitGUID(unitID), unitID)
	end
end

local function FindPlateWithRaidIcon(iconNum)
	for frame in pairs(lib.nameplates) do 
		if lib:IsMarked(frame) and lib:GetRaidIcon(frame) == iconNum then
			return frame
		end
	end
	return nil
end

local function FindGUIDByRaidIcon(frame, raidNum, from)
	local group, num = "", 0
	if GetNumRaidMembers() > 1 then
		group, num = "raid", GetNumRaidMembers()
	elseif GetNumPartyMembers() > 0 then
		group, num = "party", GetNumPartyMembers()
	else
		return
	end
	
	local unitID, targetID, targetIndex
	for i = 1, num do
		unitID = group..i;
		targetID = unitID.."target"
		if UnitExists(targetID) and not UnitIsUnit("target", targetID) then
			targetIndex = GetRaidTargetIndex(targetID)
			if targetIndex and targetIndex == raidNum then
				debugPrint("FindGUIDByRaidIcon", from, "Icon:"..tostring(raidNum), "unitID:"..tostring(targetID), "GUID:"..tostring(UnitGUID(targetID)))
				FoundPlateGUID(frame, UnitGUID(targetID), targetID)
				return
			end
		end
	end
end

local function UpdateNameplateInfo(frame)
	if lib:IsMouseover(frame) and not UnitExists("mouseover") then
		HideMouseoverRegion(frame)
	end

	if not lib.plateGUIDs[frame] then
		if lib:IsMouseover(frame) then
			GetMouseoverGUID(frame)
		elseif lib:IsMarked(frame) then
			local raidNum = lib:GetRaidIcon(frame)
			if raidNum and raidNum > 0 then
				FindGUIDByRaidIcon(frame, raidNum, "UpdateNameplateInfo")
			end
		end
	end

--~ 	debugPrint("Updated nameplate", lib:GetName(frame))
	frame.lnpLastUpdate = 0
end

function lib.OnNameplateShow(self, ...)
	lib:SetupNameplate(self)
end
local function ourOnShow(...) return lib.OnNameplateShow(...) end

function lib.OnNameplateHide(self, ...)
	lib.isOnScreen[self] = false
	RecycleNameplate(self)
end
local function ourOnHide(...) return lib.OnNameplateHide(...) end

--[[
local function ourOnHide(self, ...)
	lib.isOnScreen[self] = false
	RecycleNameplate(self)
end
]]

local callbackPlateChange = {}
local watchForPlateChange = false --Only watch for visible plate change if a addon has registered the callback.

function lib.callbacks:OnUsed(target, eventname)
--~ 	if eventname == "LibNameplate_VisiblePlateChange" then
--~ 		table_insert(callbackPlateChange, target)
--~ 		watchForPlateChange = true
--~ 	end
end

function lib.callbacks:OnUnused(target, eventname)
--~ 	if eventname == "LibNameplate_VisiblePlateChange" then
--~ 		for i=table_getn(callbackPlateChange),1, -1  do 
--~ 			if callbackPlateChange[i] == target then
--~ 				table_remove(callbackPlateChange, i)
--~ 				break
--~ 			end
--~ 		end
--~ 		if table_getn(callbackPlateChange) == 0 then
--~ 			watchForPlateChange = false
--~ 		end
--~ 	end
end

function lib.OnNameplateUpdate(self, elapsed, ...)
	if self.lnpCheckForTarget then --Check on the first OnUpdate after the frame's shown.
		self.lnpCheckForTarget = false
		if not lib.plateGUIDs[self] and self:IsShown() and ((self:GetAlpha() == 1) and UnitExists("target")) then
			FoundPlateGUID(self, UnitGUID("target"), "target")
		end
	end

	self.lnpLastUpdate = (self.lnpLastUpdate or 0) + elapsed
	if self.lnpLastUpdate > updateDelay then
		UpdateNameplateInfo(self)
		
--~ 		if watchForPlateChange and not fakePlate[self] then
--~ 			if self.extended then
--~ 				lib.callbacks:Fire("LibNameplate_VisiblePlateChange", self, self.extended)
--~ 				fakePlate[self] = true
--~ 			end
--~ 		end

		if not fakePlate[self] and self.extended then
			realPlate[self.extended] = self
			fakePlate[self] = self.extended

			lib.callbacks:Fire("LibNameplate_RecycleNameplate", self)--Hide real plate so addon unhook their stuff.
			lib.callbacks:Fire("LibNameplate_NewNameplate", fakePlate[self])

		end
	end
end
local function ourOnUpdate(...) return lib.OnNameplateUpdate(...) end

function lib:SetupNameplate(frame)
	self.isOnScreen[frame] = true

	if frame:HasScript("OnHide") and not lib.onHideHooks[frame] then
		lib.onHideHooks[frame] = true
		frame:HookScript("OnHide", ourOnHide)
	end
	
	if frame:HasScript("OnShow") and not lib.onShowHooks[frame] then
		lib.onShowHooks[frame] = true
		frame:HookScript("OnShow", ourOnShow)
	end

	if frame:HasScript("OnUpdate") and not lib.onUpdateHooks[frame] then
		lib.onUpdateHooks[frame] = true
		frame:HookScript("OnUpdate", ourOnUpdate)
	end
	
	local plateName = self:GetName(frame)
--~ 	debugPrint("SetupNameplate", frame, plateName)
	lib.nameplates[frame] = plateName
	lib.names[plateName] = frame
	
	-- and frame.extended:IsShown()
	if frame.extended and not fakePlate[frame] then
		fakePlate[frame] = frame.extended
		realPlate[frame.extended] = frame
		
		--Without this was causing problems with PlateBuffs where it was never told the real plate was gone.
		self.callbacks:Fire("LibNameplate_RecycleNameplate", frame)
	end
	
	lib.callbacks:Fire("LibNameplate_NewNameplate", fakePlate[frame] or frame)

	frame.lnpCheckForTarget = true
	UpdateNameplateInfo(frame)
end--	/lnbug

local function CheckForTargetGUID()
	local unitID = "target"
	local GUID
	for frame in pairs(lib.nameplates) do 
		if lib:IsTarget(frame) then
			lib.targeted = frame
			if not lib.plateGUIDs[frame] then
				FoundPlateGUID(frame, UnitGUID(unitID), unitID)
			end
			lib.callbacks:Fire("LibNameplate_TargetNameplate", fakePlate[frame] or frame)
			return
		end
	end
--~ 	debugPrint("Failed to find target GUID")
end

local function OnUpdate(self, elapsed)
	if self.checkTarget then
		self.checkTarget = false
		CheckForTargetGUID()
	end

	FindNameplates()

	self.lastUpdate = self.lastUpdate + elapsed
	if self.lastUpdate >= scanDelay then
		self.lastUpdate = 0
		for frame, value in pairs(lib.isOnScreen) do 
			if (value == true and not frame:IsShown()) then --OnHide fail
				debugPrint("OnHide fail", frame, value, frame:IsShown())
				lib.onHideHooks[frame] = false
				lib.isOnScreen[frame] = nil
				lib.OnNameplateHide(frame)
				
			elseif (value == false and frame:IsShown()) then --OnShow fail
				debugPrint("OnShow fail", frame, value, frame:IsShown())
				lib.onShowHooks[frame] = false
				lib.isOnScreen[frame] = nil
				lib:SetupNameplate(frame, true)
				
			end
		end
	end
end


local function OnEvent(self, event, ...)
	if event == "UPDATE_MOUSEOVER_UNIT" then
		if GetMouseFocus():GetName() == "WorldFrame" then
			local i = 0
			local mouseoverPlate
			for frame in pairs(lib.nameplates) do 
				if frame:IsShown() and lib:IsMouseover(frame) then
					i = i + 1
					mouseoverPlate = frame

				end
			end
			if i == 1 then
				if not lib.plateGUIDs[mouseoverPlate] then
					GetMouseoverGUID(mouseoverPlate)
				end
				lib.callbacks:Fire("LibNameplate_MouseoverNameplate", fakePlate[mouseoverPlate] or mouseoverPlate)
			elseif i > 1 then
				debugPrint(i.." mouseover frames")
			end
			
		end
		
	elseif event == "PLAYER_TARGET_CHANGED" then
		if UnitExists("target") then
			self.checkTarget = true--Check for target nameplate on next OnUpdate. We can't do it now.
		else
			lib.targeted = nil
		end
		
	elseif event == "UNIT_TARGET" then
		local unitID = ...
		local targetID = unitID.."target"
		if UnitExists(targetID) and not UnitIsUnit("player", unitID) then
			local iconNum = GetRaidTargetIndex(targetID)
			if iconNum and iconNum > 0 then
				local foundPlate = FindPlateWithRaidIcon(iconNum)
				if foundPlate and not lib.plateGUIDs[foundPlate] then
--~ 					debugPrint(event, "Found raid icon on ", UnitName(unitID), "'s target", UnitName(targetID), "icon:"..tostring(iconNum))
					FoundPlateGUID(foundPlate, UnitGUID(targetID), targetID)
				end
			end
		end
		
	elseif event == "RAID_TARGET_UPDATE" then
		for frame in pairs(lib.nameplates) do 
			if frame:IsShown() and not lib.plateGUIDs[frame] and lib:IsMarked(frame) then
				local raidNum = lib:GetRaidIcon(frame)
				if raidNum and raidNum > 0 then
					FindGUIDByRaidIcon(frame, raidNum, event)
				end
			end
		end
	elseif event == "PLAYER_ENTERING_WORLD" then
--~ 		debugPrint(event, MAJOR, _G.TidyPlates, _G.Aloft)
		if _G.TidyPlates then
			tidyPlatesRunning = true
		elseif _G.Aloft then
			aloftRunning = true
		elseif IsAddOnLoaded("caelNamePlates") then
			caelNPRunning = true
			scanDelay = 0.1 --cael breaks our onhide/onshow hooks. 
--~ 			debugPrint("caelNamePlates")
		end
		
	end
end


if not lib.frame then
	lib.frame=CreateFrame("Frame")
end

lib.frame.lastUpdate = 0
lib.frame.checkTarget = false
lib.frame:SetScript("OnUpdate", OnUpdate)
lib.frame:SetScript("OnEvent", OnEvent)
lib.frame:RegisterEvent('UPDATE_MOUSEOVER_UNIT')
lib.frame:RegisterEvent('PLAYER_TARGET_CHANGED')
lib.frame:RegisterEvent('UNIT_TARGET')
lib.frame:RegisterEvent('RAID_TARGET_UPDATE')
lib.frame:RegisterEvent('PLAYER_ENTERING_WORLD')